@@ -1,6 +1,6 @@
 /* dirmngr.c - Keyserver and X.509 LDAP access
  * Copyright (C) 2002 Klarälvdalens Datakonsult AB
- * Copyright (C) 2003, 2004, 2006, 2007, 2008, 2010, 2011 g10 Code GmbH
+ * Copyright (C) 2003, 2004, 2006, 2007, 2008, 2010, 2011, 2020 g10 Code GmbH
  * Copyright (C) 2014 Werner Koch
  *
  * This file is part of GnuPG.
@@ -17,8 +17,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- *
- * SPDX-License-Identifier: GPL-3.0+
+ * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -100,6 +99,7 @@ enum cmd_and_opt_values {
   aFlush,
   aGPGConfList,
   aGPGConfTest,
+  aGPGConfVersions,
 
   oOptions,
   oDebug,
@@ -124,6 +124,7 @@ enum cmd_and_opt_values {
   oHTTPProxy,
   oLDAPProxy,
   oOnlyLDAPProxy,
+  oLDAPServer,
   oLDAPFile,
   oLDAPTimeout,
   oLDAPAddServers,
@@ -174,6 +175,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aFlush,    "flush",     N_("flush the cache")),
   ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
   ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
+  ARGPARSE_c (aGPGConfVersions, "gpgconf-versions", "@"),
 
   ARGPARSE_group (301, N_("@\nOptions:\n ")),
 
@@ -181,7 +183,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oQuiet,    "quiet",     N_("be somewhat more quiet")),
   ARGPARSE_s_n (oSh,       "sh",        N_("sh-style command output")),
   ARGPARSE_s_n (oCsh,      "csh",       N_("csh-style command output")),
-  ARGPARSE_s_s (oOptions,  "options",   N_("|FILE|read options from FILE")),
+  ARGPARSE_conffile (oOptions, "options", N_("|FILE|read options from FILE")),
   ARGPARSE_s_s (oDebugLevel, "debug-level",
                 N_("|LEVEL|set the debugging level to LEVEL")),
   ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
@@ -208,6 +210,8 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oOnlyLDAPProxy, "only-ldap-proxy",
                 N_("do not use fallback hosts with --ldap-proxy")),
 
+  ARGPARSE_s_s (oLDAPServer, "ldapserver",
+                N_("|SPEC|use this keyserver to lookup keys")),
   ARGPARSE_s_s (oLDAPFile, "ldapserverlist-file",
                 N_("|FILE|read LDAP server list from FILE")),
   ARGPARSE_s_n (oLDAPAddServers, "add-servers",
@@ -259,6 +263,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oConnectTimeout, "connect-timeout", "@"),
   ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
   ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
+  ARGPARSE_noconffile (oNoOptions, "no-options", "@"),
 
   ARGPARSE_group (302,N_("@\n(See the \"info\" manual for a complete listing "
                          "of all commands and options)\n")),
@@ -342,6 +347,11 @@ static int network_activity_seen;
 /* A list of filenames registred with --hkp-cacert.  */
 static strlist_t hkp_cacert_filenames;
 
+/* A flag used to clear the list of ldapservers iff --ldapserver is
+ * given on the command line or one of the conf files. In this case we
+ * want to clear all old specifications through the legacy
+ * dirmngr_ldapservers.conf. */
+static int ldapserver_list_needs_reset;
 
 /* The timer tick used for housekeeping stuff.  The second constant is used when a shutdown is pending.  */
 #define TIMERTICK_INTERVAL           (60)
@@ -378,6 +388,8 @@ static ldap_server_t parse_ldapserver_file (const char* filename, int ienoent);
 static fingerprint_list_t parse_ocsp_signer (const char *string);
 static void netactivity_action (void);
 static void handle_connections (assuan_fd_t listen_fd);
+static void gpgconf_versions (void);
+
 
 /* NPth wrapper function definitions. */
 ASSUAN_SYSTEM_NPTH_IMPL;
@@ -388,9 +400,11 @@ my_strusage( int level )
   const char *p;
   switch ( level )
     {
+    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "@DIRMNGR@ (@GNUPG@)";
       break;
     case 13: p = VERSION; break;
+    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
       /* TRANSLATORS: @EMAIL@ will get replaced by the actual bug
          reporting address.  This is so that we can change the
@@ -631,6 +645,8 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       set_dns_timeout (0);
       opt.connect_timeout = 0;
       opt.connect_quick_timeout = 0;
+      opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
+      ldapserver_list_needs_reset = 1;
       return 1;
     }
 
@@ -719,6 +735,30 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
     case oStandardResolver: enable_standard_resolver (1); break;
     case oRecursiveResolver: enable_recursive_resolver (1); break;
 
+    case oLDAPServer:
+      {
+        ldap_server_t server;
+        char *p;
+
+        p = pargs->r.ret_str;
+        if (!strncmp (p, "ldap:", 5) && !(p[5] == '/' && p[6] == '/'))
+          p += 5;
+
+        server = ldapserver_parse_one (p, NULL, 0);
+        if (server)
+          {
+            if (ldapserver_list_needs_reset)
+              {
+                ldapserver_list_needs_reset = 0;
+                ldapserver_list_free (opt.ldapservers);
+                opt.ldapservers = NULL;
+              }
+            server->next = opt.ldapservers;
+            opt.ldapservers = server;
+          }
+      }
+      break;
+
     case oKeyServer:
       if (*pargs->r.ret_str)
         add_to_strlist (&opt.keyserver, pargs->r.ret_str);
@@ -740,6 +780,10 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.connect_quick_timeout = pargs->r.ret_ulong * 1000;
       break;
 
+    case oLDAPTimeout:
+      opt.ldaptimeout = pargs->r.ret_int;
+      break;
+
     default:
       return 0; /* Not handled. */
     }
@@ -824,12 +868,10 @@ main (int argc, char **argv)
   ARGPARSE_ARGS pargs;
   int orig_argc;
   char **orig_argv;
-  FILE *configfp = NULL;
-  char *configname = NULL;
+  char *last_configname = NULL;
+  const char *configname = NULL;
   const char *shell;
-  unsigned configlineno;
-  int parse_debug = 0;
-  int default_config =1;
+  int debug_argparser = 0;
   int greeting = 0;
   int nogreeting = 0;
   int nodetach = 0;
@@ -911,63 +953,56 @@ main (int argc, char **argv)
   orig_argv = argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= 1|(1<<6);  /* do not remove the args, ignore version */
-  while (arg_parse( &pargs, opts))
+  pargs.flags= (ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
+  while (gnupg_argparse (NULL, &pargs, opts))
     {
-      if (pargs.r_opt == oDebug || pargs.r_opt == oDebugAll)
-        parse_debug++;
-      else if (pargs.r_opt == oOptions)
-        { /* Yes there is one, so we do not try the default one, but
-	     read the option file when it is encountered at the
-	     commandline */
-          default_config = 0;
-	}
-      else if (pargs.r_opt == oNoOptions)
-        default_config = 0; /* --no-options */
-      else if (pargs.r_opt == oHomedir)
+      switch (pargs.r_opt)
         {
+        case oDebug:
+        case oDebugAll:
+          debug_argparser++;
+          break;
+        case oHomedir:
           gnupg_set_homedir (pargs.r.ret_str);
+          break;
         }
     }
+  /* Reset the flags.  */
+  pargs.flags &= ~(ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
 
   socket_name = dirmngr_socket_name ();
-  if (default_config)
-    configname = make_filename (gnupg_homedir (), DIRMNGR_NAME".conf", NULL );
 
+  /* The configuraton directories for use by gpgrt_argparser.  */
+  gnupg_set_confdir (GNUPG_CONFDIR_SYS, gnupg_sysconfdir ());
+  gnupg_set_confdir (GNUPG_CONFDIR_USER, gnupg_homedir ());
+
+  /* We are re-using the struct, thus the reset flag.  We OR the
+   * flags so that the internal intialized flag won't be cleared. */
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= 1;  /* do not remove the args */
- next_pass:
-  if (configname)
+  pargs.flags |=  (ARGPARSE_FLAG_RESET
+                   | ARGPARSE_FLAG_KEEP
+                   | ARGPARSE_FLAG_SYS
+                   | ARGPARSE_FLAG_USER);
+  while (gnupg_argparser (&pargs, opts, DIRMNGR_NAME EXTSEP_S "conf"))
     {
-      configlineno = 0;
-      configfp = fopen (configname, "r");
-      if (!configfp)
+      if (pargs.r_opt == ARGPARSE_CONFFILE)
         {
-          if (default_config)
+          if (debug_argparser)
+            log_info (_("reading options from '%s'\n"),
+                      pargs.r_type? pargs.r.ret_str: "[cmdline]");
+          if (pargs.r_type)
             {
-              if( parse_debug )
-                log_info (_("Note: no default option file '%s'\n"),
-                          configname );
-	    }
+              xfree (last_configname);
+              last_configname = xstrdup (pargs.r.ret_str);
+              configname = last_configname;
+            }
           else
-            {
-              log_error (_("option file '%s': %s\n"),
-                         configname, strerror(errno) );
-              exit(2);
-	    }
-          xfree (configname);
-          configname = NULL;
-	}
-      if (parse_debug && configname )
-        log_info (_("reading options from '%s'\n"), configname );
-      default_config = 0;
-    }
-
-  while (optfile_parse( configfp, configname, &configlineno, &pargs, opts) )
-    {
+            configname = NULL;
+          continue;
+        }
       if (parse_rereadable_options (&pargs, 0))
         continue; /* Already handled */
       switch (pargs.r_opt)
@@ -982,6 +1017,7 @@ main (int argc, char **argv)
         case aFetchCRL:
 	case aGPGConfList:
 	case aGPGConfTest:
+	case aGPGConfVersions:
           cmd = pargs.r_opt;
           break;
 
@@ -991,18 +1027,8 @@ main (int argc, char **argv)
 
         case oDebugWait: debug_wait = pargs.r.ret_int; break;
 
-        case oOptions:
-          /* Config files may not be nested (silently ignore them) */
-          if (!configfp)
-            {
-		xfree(configname);
-		configname = xstrdup(pargs.r.ret_str);
-		goto next_pass;
-	    }
-          break;
         case oNoGreeting: nogreeting = 1; break;
         case oNoVerbose: opt.verbose = 0; break;
-        case oNoOptions: break; /* no-options */
         case oHomedir: /* Ignore this option here. */; break;
         case oNoDetach: nodetach = 1; break;
         case oLogFile: logfile = pargs.r.ret_str; break;
@@ -1014,9 +1040,6 @@ main (int argc, char **argv)
 #        endif /*USE_LDAP*/
           break;
 	case oLDAPAddServers: opt.add_new_ldapservers = 1; break;
-	case oLDAPTimeout:
-	  opt.ldaptimeout = pargs.r.ret_int;
-	  break;
 
         case oFakedSystemTime:
           gnupg_set_time ((time_t)pargs.r.ret_ulong, 0);
@@ -1030,20 +1053,26 @@ main (int argc, char **argv)
           listen_backlog = pargs.r.ret_int;
           break;
 
-        default : pargs.err = configfp? 1:2; break;
+        default:
+          if (configname)
+            pargs.err = ARGPARSE_PRINT_WARNING;
+          else
+            pargs.err = ARGPARSE_PRINT_ERROR;
+          break;
 	}
     }
-  if (configfp)
+  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+
+  if (!last_configname)
+    opt.config_filename = make_filename (gnupg_homedir (),
+                                         DIRMNGR_NAME EXTSEP_S "conf",
+                                         NULL);
+  else
     {
-      fclose (configfp);
-      configfp = NULL;
-      /* Keep a copy of the name so that it can be read on SIGHUP. */
-      opt.config_filename = configname;
-      configname = NULL;
-      goto next_pass;
+      opt.config_filename = last_configname;
+      last_configname = NULL;
     }
-  xfree (configname);
-  configname = NULL;
+
   if (log_get_errorcount(0))
     exit(2);
   if (nogreeting )
@@ -1073,7 +1102,7 @@ main (int argc, char **argv)
           log_info (_("Note: '%s' is not considered an option\n"), argv[i]);
     }
 
-  if (!access ("/etc/"DIRMNGR_NAME, F_OK)
+  if (!gnupg_access ("/etc/"DIRMNGR_NAME, F_OK)
       && !strncmp (gnupg_homedir (), "/etc/", 5))
     log_info
       ("NOTE: DirMngr is now a proper part of %s.  The configuration and"
@@ -1093,9 +1122,11 @@ main (int argc, char **argv)
 
   post_option_parsing ();
 
-  /* Get LDAP server list from file. */
+  /* Get LDAP server list from file unless --ldapserver has been used.  */
 #if USE_LDAP
-  if (!ldapfile)
+  if (opt.ldapservers)
+    ;
+  else if (!ldapfile)
     {
       ldapfile = make_filename (gnupg_homedir (),
                                 "dirmngr_ldapservers.conf",
@@ -1272,7 +1303,7 @@ main (int argc, char **argv)
         {
           log_error (_("error binding socket to '%s': %s\n"),
                      serv_addr.sun_path,
-                     gpg_strerror (gpg_error_from_errno (errno)));
+                     gpg_strerror (gpg_error_from_syserror ()));
           assuan_sock_close (fd);
           dirmngr_exit (1);
         }
@@ -1474,12 +1505,6 @@ main (int argc, char **argv)
       char *filename;
       char *filename_esc;
 
-      /* First the configuration file.  This is not an option, but it
-	 is vital information for GPG Conf.  */
-      if (!opt.config_filename)
-        opt.config_filename = make_filename (gnupg_homedir (),
-                                             "dirmngr.conf", NULL );
-
       filename = percent_escape (opt.config_filename, NULL);
       es_printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
               GC_OPT_FLAG_DEFAULT, filename);
@@ -1538,11 +1563,13 @@ main (int argc, char **argv)
                  filename_esc);
       xfree (filename_esc);
 
-
       es_printf ("nameserver:%lu:\n", flags | GC_OPT_FLAG_NONE);
       es_printf ("resolver-timeout:%lu:%u\n",
                  flags | GC_OPT_FLAG_DEFAULT, 0);
     }
+  else if (cmd == aGPGConfVersions)
+    gpgconf_versions ();
+
   cleanup ();
   return !!rc;
 }
@@ -1825,36 +1852,39 @@ static void
 reread_configuration (void)
 {
   ARGPARSE_ARGS pargs;
-  FILE *fp;
-  unsigned int configlineno = 0;
+  char *twopart;
   int dummy;
 
   if (!opt.config_filename)
     return; /* No config file. */
 
-  fp = fopen (opt.config_filename, "r");
-  if (!fp)
-    {
-      log_error (_("option file '%s': %s\n"),
-                 opt.config_filename, strerror(errno) );
-      return;
-    }
+  twopart = strconcat (DIRMNGR_NAME EXTSEP_S "conf" PATHSEP_S,
+                       opt.config_filename, NULL);
+  if (!twopart)
+    return;  /* Out of core.  */
 
   parse_rereadable_options (NULL, 1); /* Start from the default values. */
 
   memset (&pargs, 0, sizeof pargs);
   dummy = 0;
   pargs.argc = &dummy;
-  pargs.flags = 1;  /* do not remove the args */
-  while (optfile_parse (fp, opt.config_filename, &configlineno, &pargs, opts) )
+  pargs.flags = (ARGPARSE_FLAG_KEEP
+                 |ARGPARSE_FLAG_SYS
+                 |ARGPARSE_FLAG_USER);
+  while (gnupg_argparser (&pargs, opts, twopart))
     {
-      if (pargs.r_opt < -1)
-        pargs.err = 1; /* Print a warning. */
+      if (pargs.r_opt == ARGPARSE_CONFFILE)
+        {
+          log_info (_("reading options from '%s'\n"),
+                    pargs.r_type? pargs.r.ret_str: "[cmdline]");
+        }
+      else if (pargs.r_opt < -1)
+        pargs.err = ARGPARSE_PRINT_WARNING;
       else /* Try to parse this option - ignore unchangeable ones. */
         parse_rereadable_options (&pargs, 1);
     }
-  fclose (fp);
-
+  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+  xfree (twopart);
   post_option_parsing ();
 }
 
@@ -2036,7 +2066,7 @@ handle_tick (void)
 
   /* Check whether the homedir is still available.  */
   if (!shutdown_pending
-      && stat (gnupg_homedir (), &statbuf) && errno == ENOENT)
+      && gnupg_stat (gnupg_homedir (), &statbuf) && errno == ENOENT)
     {
       shutdown_pending = 1;
       log_info ("homedir has been removed - shutting down\n");
@@ -2347,3 +2377,62 @@ dirmngr_get_current_socket_name (void)
   else
     return dirmngr_socket_name ();
 }
+
+
+
+/* Parse the revision part from the extended version blurb.  */
+static const char *
+get_revision_from_blurb (const char *blurb, int *r_len)
+{
+  const char *s = blurb? blurb : "";
+  int n;
+
+  for (; *s; s++)
+    if (*s == '\n' && s[1] == '(')
+      break;
+  if (*s)
+    {
+      s += 2;
+      for (n=0; s[n] && s[n] != ' '; n++)
+        ;
+    }
+  else
+    {
+      s = "?";
+      n = 1;
+    }
+  *r_len = n;
+  return s;
+}
+
+
+/* Print versions of dirmngr and used libraries.  This is used by
+ * "gpgconf --show-versions" so that there is no need to link gpgconf
+ * against all these libraries.  This is an internal API and should
+ * not be relied upon.  */
+static void
+gpgconf_versions (void)
+{
+  const char *s;
+  int n;
+
+  /* Unfortunately Npth has no way to get the version.  */
+
+  s = get_revision_from_blurb (assuan_check_version ("\x01\x01"), &n);
+  es_fprintf (es_stdout, "* Libassuan %s (%.*s)\n\n",
+              assuan_check_version (NULL), n, s);
+
+  s = get_revision_from_blurb (ksba_check_version ("\x01\x01"), &n);
+  es_fprintf (es_stdout, "* KSBA %s (%.*s)\n\n",
+              ksba_check_version (NULL), n, s);
+
+#ifdef HTTP_USE_NTBTLS
+  s = get_revision_from_blurb (ntbtls_check_version ("\x01\x01"), &n);
+  es_fprintf (es_stdout, "* NTBTLS %s (%.*s)\n\n",
+              ntbtls_check_version (NULL), n, s);
+#elif HTTP_USE_GNUTLS
+  es_fprintf (es_stdout, "* GNUTLS %s\n\n",
+              gnutls_check_version (NULL));
+#endif
+
+}
